import os, sys, argparse, json, datetime, re

import boto

from s3ts.treestore import TreeStore, TreeStoreConfig
from s3ts.filestore import FileStore, LocalFileStore
from s3ts.s3filestore import S3FileStore
from s3ts.package import PackageJS, packageDiff, packageFilter
from s3ts.metapackage import MetaPackage, SubPackage, MetaPackageJS

def getEnv( name, desc ):
    try:
        return os.environ[name]
    except KeyError:
        sys.stderr.write( 'The environment variable {0} must be set to {1}\n'.format( name, desc ) )
        sys.exit(1)

class UploadProgress(object):
    def __init__( self ):
        self.cumTransferred = 0
        self.cumCached = 0

    def __call__( self, nDownloaded, nCached ):
        self.cumTransferred += nDownloaded
        self.cumCached += nCached
        sys.stdout.write( "\r{0} transferred + {1} cached".format( self.cumTransferred, self.cumCached ) )
        sys.stdout.flush()

class DownloadProgress(object):
    def __init__( self, pkg ):
        self.size = pkg.size()
        self.cumTransferred = 0
        self.cumCached = 0

    def __call__( self, nDownloaded, nCached ):
        self.cumTransferred += nDownloaded
        self.cumCached += nCached
        sys.stdout.write( "\r{0} transferred + {1} cached / {2} total".format( self.cumTransferred, self.cumCached, self.size ) )
        sys.stdout.flush()

class InstallProgress(object):
    def __init__( self, pkg ):
        self.size = pkg.size()
        self.cumSize = 0

    def __call__( self, nbytes ):
        self.cumSize += nbytes
        sys.stdout.write( "\r{0} / {1} installed".format( self.cumSize, self.size ) )
        sys.stdout.flush()

def outVerbose( formatStr, *args ):
    sys.stdout.write( "\r" + formatStr.format(*args) + "\n" )
    sys.stdout.flush()

def connectToBucket():
    awsAccessKeyId = getEnv( 'AWS_ACCESS_KEY_ID', 'the AWS access key id' )
    awsSecretAccessKey = getEnv( 'AWS_SECRET_ACCESS_KEY', 'the AWS secret access key' )
    bucketName = getEnv( 'S3TS_BUCKET', 'the AWS S3 bucket used for tree storage'  )
    s3PathPrefix = os.environ.get( 'S3TS_S3PREFIX' )
    s3c = boto.connect_s3(awsAccessKeyId,awsSecretAccessKey)
    return s3c.get_bucket( bucketName ),s3PathPrefix

def createTreeStore(chunksize):
    localCacheDir = getEnv( 'S3TS_LOCALCACHE', 'the local directory used for caching'  )
    bucket,s3PathPrefix = connectToBucket()
    config = TreeStoreConfig( chunksize, True )
    return TreeStore.create( S3FileStore(bucket,s3PathPrefix), LocalFileStore(localCacheDir), config )

def openTreeStore(dryRun=False,verbose=False):
    localCacheDir = getEnv( 'S3TS_LOCALCACHE', 'the local directory used for caching'  )
    bucket,s3PathPrefix = connectToBucket()
    treeStore = TreeStore.open( S3FileStore(bucket,s3PathPrefix), LocalFileStore(localCacheDir) )
    treeStore.setDryRun(dryRun)
    if verbose:
        treeStore.setOutVerbose( outVerbose )
    return treeStore

def nonS3TreeStore():
    # Don't use or require S3 - some operations won't be available
    localCacheDir = getEnv( 'S3TS_LOCALCACHE', 'the local directory used for caching'  )
    return TreeStore( FileStore(), LocalFileStore(localCacheDir), None )

def readPackageFile( packageFile ):
    with open( packageFile, 'r' ) as f:
        return PackageJS().fromJson( json.loads( f.read() ) )

def init( chunksize ):
    treeStore = createTreeStore(chunksize)

def list():
    treeStore = openTreeStore()
    for treeName in treeStore.listPackages():
        print treeName
    for treeName in treeStore.listMetaPackages():
        print treeName

def remove( treename, confirmed ):
    if not confirmed:
        s = raw_input( "Really remove tree '{}' (Y/N) ? [N] ".format(treename) )
        confirmed = s =='Y'
    if confirmed:
        print "Removing {}".format( treename )
        treestore = openTreeStore()
        treestore.remove( treename )
    else:
        print "Cancelled"

def rename( fromtreename, totreename ):
    treestore = openTreeStore()
    treestore.rename( fromtreename, totreename )

def info( treename, pathRegex ):
    treeStore = openTreeStore()
    try:
        pkg = treeStore.findMetaPackage(treename)
        infofn = printMetaPackageInfo
    except KeyError:
        pkg = treeStore.findPackage(treename)
        pkg = packageFilter(pkg,pathRegex)
        infofn = printPackageInfo
    infofn(treename, pkg)

def printPackageInfo(name, pkg):
    print 'Package:', name
    print 'Created At:', pkg.creationTime.isoformat()
    print 'Total Size: {0} bytes'.format(pkg.size())
    print 'Files:'
    for pf in pkg.files:
        print '    {0} ({1} chunks, {2} bytes)'.format( pf.path, len(pf.chunks), pf.size() )

def printMetaPackageInfo(name, pkg):
    print 'Package:', name
    print 'Created At:', pkg.creationTime.isoformat()
    print 'Components:'
    for component in pkg.components:
        print '    ', component.info()

def upload( treename, description, localdir, dryRun, verbose ):
    creationTime = datetime.datetime.now()
    treeStore = openTreeStore(dryRun=dryRun,verbose=verbose)
    treeStore.upload( treename, description, creationTime, localdir, UploadProgress() )
    print

def uploadMany( treename, description, localdir, kioskDir ):
    creationTime = datetime.datetime.now()
    treeStore = openTreeStore()
    treeStore.uploadMany(treename, description, creationTime, localdir, kioskDir, UploadProgress())
    print

def createMerged( treename, packageArgs, dryRun, verbose ):
    creationTime = datetime.datetime.now()
    treeStore = openTreeStore(dryRun=dryRun,verbose=verbose)
    packageMap = {}

    # Convert the package command line arguments from form
    #    [ "subdir1:pname1", "subdir2:pname2", ... ]
    # to
    #    { "subdir1":"pname1", "subdir2":"pname2", ... }
    
    for s in packageArgs:
        fields = s.split( ':', 2 )
        if len(fields) != 2:
            sys.stderr.write( 'create-merge arguments must be of form SUBDIR:TREENAME' )
            sys.exit(1)
        packageMap[fields[0]] = fields[1]
    
    treeStore.createMerged( treename, creationTime, packageMap)
    print                

def download( treename, dryRun, verbose, metadata ):
    treeStore = openTreeStore(dryRun=dryRun,verbose=verbose)
    pkg = treeStore.find( treename, metadata )
    treeStore.download( pkg, DownloadProgress(pkg) )
    print

def flush( dryRun, verbose ):
    treeStore = openTreeStore(dryRun=dryRun,verbose=verbose)
    treeStore.flushStore()
    print

def flushCache( dryRun, verbose, packageNames ):
    treeStore = openTreeStore(dryRun=dryRun,verbose=verbose)
    treeStore.flushLocalCache(packageNames)
    print
    
def install( treename, localdir, verbose, pathRegex, metadata ):
    treeStore = openTreeStore(verbose=verbose)
    pkg = treeStore.find( treename, metadata )
    pkg = packageFilter(pkg,pathRegex)
    treeStore.download( pkg, DownloadProgress(pkg) )
    print
    treeStore.verifyLocal( pkg )
    treeStore.install( pkg, localdir, InstallProgress(pkg) )
    print

def verifyInstall( treename, localdir, verbose, metadata ):
    treeStore = openTreeStore(verbose=verbose)
    pkg = treeStore.find( treename, metadata )
    result = treeStore.compareInstall( pkg, localdir )
    for path in result.missing:
        print "{} is missing".format(path)
    for path in result.diffs:
        print "{} is different".format(path)
    if len(result.missing) == 0 and len(result.diffs) == 0:
        print "Package {} verified ok at {}".format(treename,localdir)
    else:
        sys.exit(1)

def presign( treename, expirySecs, metadata ):
    treeStore = openTreeStore()
    pkg = treeStore.find( treename, metadata )
    treeStore.addUrls( pkg, expirySecs )
    print json.dumps( PackageJS().toJson(pkg), sort_keys=True, indent=2, separators=(',', ': ') )

def downloadHttp( packageFile ):
    treeStore = nonS3TreeStore()
    pkg = readPackageFile( packageFile )
    treeStore.downloadHttp( pkg, DownloadProgress(pkg) )
    print

def installHttp( packageFile, localdir ):
    treeStore = nonS3TreeStore()
    pkg = readPackageFile( packageFile )
    treeStore.verifyLocal( pkg )
    treeStore.install( pkg, localdir, InstallProgress(pkg) )
    print

def primeCache( localdir ):
    treeStore = openTreeStore()
    treeStore.prime( localdir, UploadProgress() )

def validateCache():
    treeStore = openTreeStore()
    print treeStore.validateLocalCache()

def comparePackages( packageName1, packageName2, metadata ):
    treeStore = openTreeStore()
    print "Fetching {}...".format(packageName1)
    package1 = treeStore.find(packageName1, metadata)
    print "Fetching {}...".format(packageName2)
    package2 = treeStore.find(packageName2, metadata)
    print "---"

    size1 = 0
    size2 = 0
    diffSize = 0

    for f in package1.files:
        size1 += f.size()
    for f in package2.files:
        size2 += f.size()
    diffPackage,removedPaths = packageDiff(package1,package2)
    for p in removedPaths:
        print "Removed", p
    for f in diffPackage.files:
        size = f.size()
        print "Updated {} (size {:,})".format(f.path,size)
        diffSize += size

    print
    print "{} size = {:,}".format(package1.name,size1)
    print "{} size = {:,}".format(package2.name,size2)
    print "update size = {:,}".format(diffSize)

TEMPLATE_METAPACKAGE_NAME =  "METANAME-VERSION"

def newMetaPackage(metapackagefile):
    template = MetaPackage(
      name = TEMPLATE_METAPACKAGE_NAME,
      description = "",
      creationTime = datetime.datetime.now(),
      components = [
        SubPackage("SUBDIR1", "PACKAGE1-VERSION"),
        SubPackage("SUBDIR2", "PACKAGE2-VERSION")
      ]
    )
    with open(metapackagefile, 'w') as f:
        f.write(json.dumps(MetaPackageJS().toJson(template), indent=2))
        f.write('\n')
    print "metapackage template written to {}".format(metapackagefile)

def uploadMetaPackage(metapackagefile):
    treeStore = openTreeStore()
    with open(metapackagefile, 'r') as f:
        metapackage = MetaPackageJS().fromJson(json.load(f))
    if metapackage.name == TEMPLATE_METAPACKAGE_NAME:
        raise RuntimeError, "Edit the metapackage template before uploading it"
    metapackage.verify(treeStore,{})
    treeStore.uploadMetaPackage(metapackage)

def downloadMetaPackage(metapackagename,metapackagefile):
    treeStore = openTreeStore()
    metapackage = treeStore.findMetaPackage(metapackagename)
    with open(metapackagefile, 'w') as f:
        f.write(json.dumps(MetaPackageJS().toJson(metapackage), indent=2))
        f.write('\n')
    
def metaDataDictionary(vals):
    """
    Turn list of colon separated pairs (from the command line) into a
    dictionary
    """
    if vals == None:
        return {}
    else:
        # 
        return dict( [v.split(":",1) for v in vals] )

def pathRegex(arg):
    if arg == None:
        return None
    else:
        return re.compile(arg)

parser = argparse.ArgumentParser()

subparsers = parser.add_subparsers(help='commands',dest='commandName')

p = subparsers.add_parser('init', help='Initialise a new store')
p.add_argument('--chunksize', action='store', default=10000000, type=int,
               help='The maximum number of bytes to be stored in each chunk')

p = subparsers.add_parser('list', help='List trees available in the store')

p = subparsers.add_parser( 'remove', help='Remove a tree from the store')
p.add_argument( "--yes", action='store_true', help='Dont ask for confirmation' )
p.add_argument('treename', action='store', help='The name of the tree')

p = subparsers.add_parser( 'rename', help='Rename an existing tree in the sore')
p.add_argument('fromtreename', action='store', help='The name of the src tree')
p.add_argument('totreename', action='store', help='The name of the target tree')

p = subparsers.add_parser('info', help='Show information about a tree')
p.add_argument('treename', action='store', help='The name of the tree')
p.add_argument('--path-regex', dest='pathRegex', action='store')

p = subparsers.add_parser('upload', help='Upload a tree from the local filesystem')
p.set_defaults(dryRun=False,verbose=False,description='')
p.add_argument('--dry-run', dest='dryRun', action='store_true')
p.add_argument('--verbose', dest='verbose', action='store_true')
p.add_argument('--description', dest='description', action='store')
p.add_argument('treename', action='store', help='The name of the tree')
p.add_argument('localdir', action='store', help='The local directory path')

p = subparsers.add_parser('download', help='Download a tree to the local cache')
p.set_defaults(dryRun=False,verbose=False)
p.add_argument('--dry-run', dest='dryRun', action='store_true')
p.add_argument('--verbose', dest='verbose', action='store_true')
p.add_argument('--meta', dest='meta', action='append')
p.add_argument('treename', action='store', help='The name of the tree')

p = subparsers.add_parser('flush', help='Flush chunks from the store that are no longer referenced')
p.set_defaults(dryRun=False,verbose=False)
p.add_argument('--dry-run', dest='dryRun', action='store_true')
p.add_argument('--verbose', dest='verbose', action='store_true')

p = subparsers.add_parser('flush-cache', help='Flush chunks from the local cache that are not referenced by the specified packages')
p.set_defaults(dryRun=False,verbose=False)
p.add_argument('--dry-run', dest='dryRun', action='store_true')
p.add_argument('--verbose', dest='verbose', action='store_true')
p.add_argument('packagenames', nargs='+' )

p = subparsers.add_parser('install', help='Download/Install a tree into the filesystem')
p.set_defaults(verbose=False)
p.add_argument('--verbose', dest='verbose', action='store_true')
p.add_argument('--meta', dest='meta', action='append')
p.add_argument('--path-regex', dest='pathRegex', action='store')
p.add_argument('treename', action='store', help='The name of the tree')
p.add_argument('localdir', action='store', help='The local directory path')

p = subparsers.add_parser('verify-install', help='Confirm a tree has been correctly installed')
p.add_argument('--verbose', dest='verbose', action='store_true')
p.add_argument('--meta', dest='meta', action='append')
p.add_argument('treename', action='store', help='The name of the tree')
p.add_argument('localdir', action='store', help='The local directory path')

p = subparsers.add_parser('presign', help='Generate a package definition containing presigned urls ')
p.add_argument('treename', action='store', help='The name of the tree')
p.add_argument('--expirySecs', action='store', default=3600, type=int,
                            help='Validity of the presigned URLs in seconds')
p.add_argument('--meta', dest='meta', action='append')

p = subparsers.add_parser('download-http', help='Download a tree to the local cache using a presigned package file')
p.add_argument('pkgfile', action='store', help='The file containing the package definition')

p = subparsers.add_parser('install-http', help='Install a tree from local cache using a presigned package file')
p.add_argument('pkgfile', action='store', help='The file containing the package definition')
p.add_argument('localdir', action='store', help='The local directory path')

p = subparsers.add_parser('prime-cache', help='Prime the local cache with the contents of a local directory')
p.add_argument('localdir', action='store', help='The local directory path')

p = subparsers.add_parser('upload-many', help='Upload multiple trees from the local filesystem')
p.set_defaults(description='')
p.add_argument('--description', dest='description', action='store')
p.add_argument('treename', action='store', help='The name of the tree')
p.add_argument('localdir', action='store', help='The local directory path')
p.add_argument('local_variant_dir', action='store', help='The local variant path')

p = subparsers.add_parser('create-merged', help='Create a new package by merging existing packages')
p.set_defaults(dryRun=False,verbose=False)
p.add_argument('--dry-run', dest='dryRun', action='store_true')
p.add_argument('--verbose', dest='verbose', action='store_true')
p.add_argument('treename', action='store', help='The name of the merged tree')
p.add_argument('package_args', nargs='+', metavar='DIR:RNAME')

p = subparsers.add_parser('new-metapackage', help='Write a template for a metapackage to a local file')
p.add_argument('metapackagefile', action='store', help='The metapackage file to which the template is to be written')

p = subparsers.add_parser('upload-metapackage', help='Upload a metapackage from a local file')
p.add_argument('metapackagefile', action='store', help='The metapackage file to be uploaded')

p = subparsers.add_parser('download-metapackage', help='Download an existing metapackage to a local file')
p.add_argument('metapackagename', action='store', help='The name of the metapackage to be downloaded')
p.add_argument('metapackagefile', action='store', help='The metapackage file to be uploaded')

p = subparsers.add_parser('compare-packages', help='Compare two packages')
p.add_argument('--meta', dest='meta', action='append')
p.add_argument('package1', action='store', help='The first package')
p.add_argument('package2', action='store', help='The second package')

validate_local_cache_parser = subparsers.add_parser('validate-local-cache', help='Validates the local cache')

def main():
    args = parser.parse_args()
    if args.commandName == 'init':
        init( args.chunksize )
    elif args.commandName == 'list':
        list()
    elif args.commandName == 'remove':
        remove( args.treename, args.yes )
    elif args.commandName == 'rename':
        rename( args.fromtreename, args.totreename )
    elif args.commandName == 'info':
        info( args.treename, pathRegex(args.pathRegex) )
    elif args.commandName == 'upload':
        upload( args.treename, args.description, args.localdir, args.dryRun, args.verbose )
    elif args.commandName == 'download':
        download( args.treename, args.dryRun, args.verbose, metaDataDictionary(args.meta) )
    elif args.commandName == 'flush':
        flush( args.dryRun, args.verbose )
    elif args.commandName == 'flush-cache':
        flushCache( args.dryRun, args.verbose, args.packagenames )
    elif args.commandName == 'install':
        install( args.treename, args.localdir, args.verbose, pathRegex(args.pathRegex), metaDataDictionary(args.meta) )
    elif args.commandName == 'verify-install':
        verifyInstall( args.treename, args.localdir, args.verbose, metaDataDictionary(args.meta) )
    elif args.commandName == 'presign':
        presign( args.treename, args.expirySecs, metaDataDictionary(args.meta) )
    elif args.commandName == 'download-http':
        downloadHttp( args.pkgfile )
    elif args.commandName == 'install-http':
        installHttp( args.pkgfile, args.localdir )
    elif args.commandName == 'prime-cache':
        primeCache( args.localdir )
    elif args.commandName == 'upload-many':
        uploadMany(args.treename, args.description, args.localdir, args.local_variant_dir)
    elif args.commandName == 'create-merged':
        createMerged(args.treename, args.package_args, args.dryRun, args.verbose)
    elif args.commandName == 'validate-local-cache':
        validateCache()
    elif args.commandName == 'compare-packages':
        comparePackages(args.package1, args.package2, metaDataDictionary(args.meta))
    elif args.commandName == 'new-metapackage':
        newMetaPackage(args.metapackagefile)
    elif args.commandName == 'upload-metapackage':
        uploadMetaPackage(args.metapackagefile)
    elif args.commandName == 'download-metapackage':
        downloadMetaPackage(args.metapackagename, args.metapackagefile)

if __name__ == '__main__':
    main()



import json, datetime, os

from s3ts.utils import datetimeFromIso

ENCODING_RAW = 'raw'
ENCODING_ZLIB = 'zlib'

class Package(object):
    """represents a collection of files to be downloaded."""
    
    def __init__( self, name, description, creationTime, files ):
        self.name = name
        self.description = description
        self.creationTime = creationTime
        self.files = files

    def size(self):
        return sum( [pf.size() for pf in self.files] )

class PackageFile(object):
    """represents a single file to be downloaded."""
    def __init__( self, sha1, path, chunks ):
        self.sha1 = sha1
        self.path = path
        self.chunks = chunks

    def size(self):
        return sum( [c.size for c in self.chunks] )

class FileChunk(object):
    """represents a chunk of a file to be downloaded"""
    def __init__( self, sha1, size, encoding, url ):
        self.sha1 = sha1
        self.size = size
        self.encoding = encoding
        self.url = url

class PackageJS(object):
    """A json de/serialiser for Package objects"""
    def __init__( self ):
        self.packageFileJS = PackageFileJS()

    def fromJson( self, jv ):
        return Package(
            jv['name'],
            jv.get('description', ''),
            datetimeFromIso( jv['creationTime'] ),
            [self.packageFileJS.fromJson(jv1) for jv1 in jv['files']]
        )
                        
    def toJson( self, v ):
        return {
            'name' : v.name,
            'description' : v.description,
            'creationTime' : v.creationTime.isoformat(),
            'files' : [ self.packageFileJS.toJson(f) for f in v.files ],
        }

class PackageFileJS(object):
    "A json de/serialiser for PackageFile objects"
    def __init__( self ):
        self.chunkJS = FileChunkJS()

    def toJson( self, v ):
        return {
            'sha1' : v.sha1,
            'path' : v.path,
            'chunks' : [ self.chunkJS.toJson(v) for v in v.chunks ]
        }

    def fromJson( self, jv ):
        return PackageFile( jv['sha1'], jv['path'], [self.chunkJS.fromJson(jv1) for jv1 in jv['chunks']] )


class FileChunkJS(object): 
    "A json de/serialiser for FileChunk objects"

    def toJson( self, v ):
        jv = {
            'sha1' : v.sha1,
            'size' : v.size,
            'encoding' : v.encoding
        }
        if v.url != None:
            jv['url'] = v.url
        return jv

    def fromJson( self, jv ):
        return FileChunk( jv['sha1'], jv['size'], jv['encoding'], jv.get('url') )

def packageFilter(package, pathRegex ):
    """
    Return a new package that only includes paths from the
    input package that match the given regexp

    """
    result = Package(
        name=package.name,
        description=package.description,
        creationTime=package.creationTime,
        files=[]
    )
    for file in package.files:
        if not pathRegex or pathRegex.match(file.path):
            result.files.append(file)
    return result

def pathFromFileSystem(fspath):
    """
    Turn a local filesystem path into a package file path.
    (Inside the package metadata, we always store paths in unix format)
    """
    return fspath.replace("\\", "/")

def packageDiff(package1, package2):
    """
    Compute a package that reflects the changes required to turn package1
    into package2. The set of file path that must be deleted is also returned.
    """

    def filesByPath(package):
        dict = {}
        for f in package.files:
            dict[f.path] = f
        return dict

    files1 = filesByPath(package1)
    files2 = filesByPath(package2)
    paths1 = set(files1.keys())
    paths2 = set(files2.keys())

    removedPaths = paths1.difference(paths2)
    commonPaths = paths1.intersection(paths2)
    addedPaths = paths2.difference(paths1)
    
    diffPackage = Package(
        name="%s->%s" % (package1.name,package2.name),
        description='',
        creationTime=package2.creationTime,
        files=[]
        )

    for p in addedPaths:
        diffPackage.files.append(files2[p])

    for p in commonPaths:
        if files1[p].sha1 != files2[p].sha1:
            diffPackage.files.append(files2[p])

    return diffPackage,removedPaths
            
S3TS_PACKAGEFILE = '.s3ts.package' 

def writeInstallPackage( installDir, pkg ):
    with open( os.path.join( installDir, S3TS_PACKAGEFILE ), 'w' ) as f:
        f.write( json.dumps( PackageJS().toJson( pkg ) ) )

def readInstallPackage( installDir ):
    with open( os.path.join( installDir, S3TS_PACKAGEFILE ), 'r' ) as f:
        return PackageJS().fromJson( json.loads( f.read() ) )



import os, hashlib, zlib, tempfile, datetime, time, shutil
import requests
            
from s3ts.config import TreeStoreConfig, TreeStoreConfigJS, InstallProperties, writeInstallProperties, S3TS_PROPERTIES
from s3ts import package, filewriter, utils, metapackage

CONFIG_PATH = 'config'
TREES_PATH = 'trees'
META_TREES_PATH = 'meta'
CHUNKS_PATH = 'chunks'
RAW_PATH = 'raw'
ZLIB_PATH = 'zlib'


class TreeStore(object):
    """implements a directory tree store

    This class implements a simple directory tree storage mechanism. It supports
    uploading and downloading of multiple directory trees, with data chunking,
    and de-duplication. A local cache is kept for download optimisation.
    """

    @classmethod
    def create( cls, pkgStore, localCache, config ):
        """creates a new treestore"""
        try:
            pkgStore.get( CONFIG_PATH )
            raise RuntimeError, "treestore is already initialised"
        except KeyError:
            pkgStore.putToJson( CONFIG_PATH, config, TreeStoreConfigJS() )
            return cls( pkgStore, localCache, config )

    @classmethod
    def open( cls, pkgStore, localCache ):
        """opens a existing treestore"""
        return cls( pkgStore, localCache, pkgStore.getFromJson( CONFIG_PATH, TreeStoreConfigJS() ) )

    @classmethod
    def forHttpOnly( cls, localCache ):
        """create a treestore that only supports the downloadHttp method"""
        return cls( None, localCache, None )

    def __init__( self, pkgStore, localCache, config ):
        self.pkgStore = pkgStore
        self.localCache = localCache
        self.config = config
        self.dryRun = False
        self.outVerbose = lambda *args : None
    def setDryRun( self, dryRun ):
        """Set the dryRun flag.

        Uploads will not push any data to S3. Downloads will
        retrieve package files, but no check data.
        """
        self.dryRun = dryRun

    def setOutVerbose( self, outVerbose ):
        """Set the function to generate verbose output

        The supplied function must take a format string, with additional
        format arguments.
        """
        self.outVerbose = outVerbose

    def upload( self, treeName, description, creationTime, localPath, progressCB ):
        """Creates a package for the content of localPath.

        This uploads the package definition and any file chunks not already uploaded.
        progressCB will be called with parameters (bytesUploaded,bytesCached) as the upload progresses

        """
        packageFiles = self.__storeFiles( self.pkgStore, localPath, progressCB )
        pkg = package.Package( treeName, description, creationTime, packageFiles )
        if not self.dryRun:
            self.outVerbose( "Uploading package definition for {}", treeName )
            self.pkgStore.putToJson( self.__treeNamePath( self.pkgStore, treeName ), pkg, package.PackageJS() )
        return pkg

    def uploadMetaPackage(self, meta):
        """ Upload a meta package to the store.
        """
        self.pkgStore.putToJson( self.__metaTreeNamePath( self.pkgStore, meta.name ), meta, metapackage.MetaPackageJS() )

    def uploadMany( self, treeName, description, creationTime, commonLocalPath, variantsLocalPath, progressCB ):
        """Creates multiple package for the content of commonPath including variantsLocalPath files
        for each directory found in the variantsLocalPath.

        This uploads the package definition and any file chunks not already uploaded.
        progressCB will be called with parameters (bytesUploaded,bytesCached) as the upload progresses

        """
        packageFiles = self.__storeFiles( self.pkgStore, commonLocalPath, progressCB )

        for variantPath in os.listdir(variantsLocalPath):
            if os.path.isdir(os.path.join(variantsLocalPath, variantPath )):
                variantTreeName = treeName + ":" + variantPath
                indivisualPackageFiles = self.__storeFiles( self.pkgStore, os.path.join( variantsLocalPath, variantPath ), progressCB )
                # merge the common + indivisual package
                mergedPackageFiles = packageFiles + indivisualPackageFiles
                pkg = package.Package( variantTreeName, description, creationTime, mergedPackageFiles)
                self.pkgStore.putToJson( self.__treeNamePath( self.pkgStore, variantTreeName ), pkg, package.PackageJS() )

    def createMerged( self, treeName, creationTime, packageMap ):
        """Create a new package by merging together existing packages.

        `packageMap` is a dictionary with subdirectories as keys, and
        package names as values.
        """
        files = []
        description = "merged package:"
        
        for subdir,subTreeName in packageMap.items():
            subpackage = self.findPackage( subTreeName )
            for file in subpackage.files:
                path = package.pathFromFileSystem( os.path.normpath( os.path.join(subdir, file.path) ) )
                files.append( package.PackageFile( file.sha1, path, file.chunks ) )
            description += "\n    {} : {} (created {})".format( subdir, subTreeName, subpackage.creationTime.isoformat() )

        pkg = package.Package( treeName, description, creationTime, files )
        if not self.dryRun:
            self.outVerbose( "Uploading package definition for {}", treeName )
            self.pkgStore.putToJson( self.__treeNamePath( self.pkgStore, treeName ), pkg, package.PackageJS() )
        return pkg
        
    def find( self, treeName, metadata ):
        """
        Return the package definition for the given name.

        Metapackages will be detected, and the equivalent
        regular package will be returned.
        """
        try:
            return self.findMetaPackage(treeName).package(self, metadata)
        except KeyError:
            return self.findPackage(treeName)

    def findPackage( self, treeName ):
        """Return the package definition for the given name"""
        return self.pkgStore.getFromJson( self.__treeNamePath( self.pkgStore, treeName ), package.PackageJS() )

    def findMetaPackage( self, metaTreeName ):
        """Return the meta package definition for the given name"""
        return self.pkgStore.getFromJson( self.__metaTreeNamePath( self.pkgStore, metaTreeName ), metapackage.MetaPackageJS() )

    def listPackages( self ):
        """Returns the available packages names"""
        return self.pkgStore.list( TREES_PATH )

    def listMetaPackages( self ):
        """Returns the available meta packages names"""
        return self.pkgStore.list( META_TREES_PATH )

    def remove( self, treeName ):
        """Removes a tree from the store"""
        return self.pkgStore.remove( self.__treeNamePath( self.pkgStore, treeName ) )

    def rename( self, fromTreeName, toTreeName ):
        """Renames a tree in the store"""
        fromPath = self.__treeNamePath( self.pkgStore, fromTreeName )
        toPath = self.__treeNamePath( self.pkgStore, toTreeName )
        pkg = self.pkgStore.getFromJson( fromPath, package.PackageJS() )
        pkg.name = toTreeName
        self.pkgStore.putToJson( toPath, pkg, package.PackageJS() )
        self.pkgStore.remove( fromPath )

    def verify( self, pkg ):
        """confirms that all data for the given package is present in the store"""
        self.__verifyStore( self.pkgStore, pkg )

    def verifyLocal( self, pkg ):
        """confirms that all data for the given package is present in the local cache"""
        self.__verifyStore( self.localCache, pkg )

    def download( self,  pkg, progressCB ):
        """downloads all data not already present to the local cache

        progressCB will be called with parameters (bytesDownloaded,bytesFromCache) as the download progresses

        """
        for pf in pkg.files:
            for chunk in pf.chunks:
                cpath = self.__chunkPath( self.pkgStore, chunk.sha1, chunk.encoding )
                lpath = self.__chunkPath( self.localCache, chunk.sha1, chunk.encoding )

                if self.localCache.exists( lpath ):
                    progressCB( 0, chunk.size )
                else:
                    if not self.dryRun:
                        self.outVerbose( "Fetching chunk {} to local cache", chunk.sha1 )
                        buf = self.pkgStore.get( cpath )
                        self.__checkSha1( self.__decompress( buf, chunk.encoding ), chunk.sha1, cpath )
                        self.localCache.put( lpath, buf )
                    progressCB( chunk.size, 0 )

    def downloadHttp( self, pkg, progressCB ):
        """downloads all data not already present to the local cache, using http.

        This requires that pkg already has embedded urls, created with the addUrls method
        progressCB will be called with parameters (bytesDownloaded,bytesFromCache) as the download progresses

        """
        for pf in pkg.files:
            for chunk in pf.chunks:
                lpath = self.__chunkPath( self.localCache, chunk.sha1, chunk.encoding )

                if self.localCache.exists( lpath ):
                    progressCB( 0, chunk.size )
                else:
                    resp = requests.get( chunk.url )
                    resp.raise_for_status()
                    buf = resp.content
                    self.__checkSha1( self.__decompress( buf, chunk.encoding ), chunk.sha1, lpath )
                    self.localCache.put( lpath, buf )
                    progressCB( chunk.size, 0 )


    def sync( self, pkg, localPath, progressCB ):
        """synchronise the content of localpath with the given package,
        reusing existing files where possible.

        progressCB will be called with parameters (nBytes) as the installation progresses,
        """
        try:
            existingPkg = package.readInstallPackage(localPath)
        except IOError:
            existingPkg = None

        if not existingPkg:
            # Start from scratch
            syncPkg,pathsToRemove = pkg,[]
            if os.path.exists( localPath ):
                shutil.rmtree( localPath )
            os.makedirs( localPath )
        else:
            # Synchronise the existing content. To ensure the
            # drectory is clean, we remove every exising file that
            # is not present in the new pkg
            syncPkg,_ = package.packageDiff( existingPkg, pkg )
            localPaths = set( utils.allFilePaths(localPath) )

            # normpath is required here to turn a package path (delimited by '/')
            # into a local filesystem path (delimited by '\' on windows)
            targetPaths = set( [os.path.normpath(f.path) for f in pkg.files] )
            pathsToRemove = localPaths.difference(targetPaths)
            pathsToRemove.discard(package.S3TS_PACKAGEFILE)
            
            # Remove the existing package from disk, so that if anything
            # fails during the sync, we start from scratch next time
            os.unlink( os.path.join( localPath, package.S3TS_PACKAGEFILE ) )

        installTime = datetime.datetime.now()
        # Remove existing files first, and any empty directories
        # to ensure that if we replace a directory with a file,
        # it's gone by the time we install it
        for path in pathsToRemove:
            path = os.path.join( localPath, path )
            self.outVerbose( "removing {}", path )
            os.unlink( path )
        utils.removeEmptyDirectories(localPath, removeRoot=False)
        self.__install( syncPkg, localPath, progressCB )
        package.writeInstallPackage( localPath, pkg )
        writeInstallProperties( localPath, InstallProperties( pkg.name, installTime ) )
            
    def install( self, pkg, localPath, progressCB ):
        """installs the given package into the local path

        progressCB will be called with parameters (nBytes) as the installation progresses,
        """
        installTime = datetime.datetime.now()
        self.__install( pkg, localPath, progressCB )
        writeInstallProperties( localPath, InstallProperties( pkg.name, installTime ) )

    def __install( self, pkg, localPath, progressCB ):

        for pf in pkg.files:
            targetPath = os.path.join( localPath, pf.path )
            targetDir = os.path.dirname(targetPath)

            if not os.path.exists( targetDir ):
                os.makedirs( targetDir )

            filesha1 = hashlib.sha1()
            # We can update the file in place, because we never install
            # to a directory tree that is in use.
            with filewriter.InPlaceFileWriter(targetPath) as f:
                for chunk in pf.chunks:
                    cpath = self.__chunkPath( self.localCache, chunk.sha1, chunk.encoding )
                    buf = self.localCache.get( cpath )
                    buf = self.__decompress( buf, chunk.encoding )
                    filesha1.update( buf )
                    f.write( buf )
                    progressCB( len(buf) )

            if filesha1.hexdigest() != pf.sha1:
                raise RuntimeError, "sha1 for {0} doesn't match".format(pf.path)

            self.outVerbose( "Wrote {}", targetPath )

    def compareInstall( self, pkg, localPath ):
        """
        Compares the package against the files installed in the given directory.
        Returns an object with 3 fields:

               result.missing - paths of files present in the package but missing on disk
               result.extra   - paths of files present on disk, but missing in the package
               result.diffs   - paths with different content
        """
        
        # 1) check that the list of files installed matches the list in the package
        installedFiles = []
        for root, dirs, files in os.walk(localPath):
            for file in files:
                installedFiles.append(os.path.relpath( os.path.join(root, file), localPath ))
        installedFiles = set(installedFiles)
        installedFiles.discard( S3TS_PROPERTIES )
        installedFiles.discard( package.S3TS_PACKAGEFILE )

        packageFiles = [os.path.normpath(f.path) for f in pkg.files]
        packageFiles = set(packageFiles)

        class Result: pass

        result = Result()
        result.missing = packageFiles.difference(installedFiles)
        result.extra = installedFiles.difference(packageFiles)
        result.diffs = set()

        # 2) verify the contents of each file
        for pf in pkg.files:
            ppath = os.path.normpath(pf.path)
            if ppath in installedFiles:
                path = os.path.join(localPath, ppath)
                filesha1 = hashlib.sha1()
                i = 0
                with open( path, 'rb' ) as f:
                    for chunk in pf.chunks:
                        buf = f.read( self.config.chunkSize )
                        chunksha1 = hashlib.sha1()
                        chunksha1.update(buf)
                        filesha1.update(buf)
                        if chunksha1.hexdigest() != chunk.sha1:
                            result.diffs.add( ppath )
                        i += len(buf)
                if filesha1.hexdigest() != pf.sha1:
                    result.diffs.add( ppath )

        return result

    def addUrls( self, pkg, expiresInSecs ):
        """Update the given package so that it can be accessed directly via pre-signed http urls"""
        for pf in pkg.files:
            for chunk in pf.chunks:
                cpath = self.__chunkPath( self.pkgStore, chunk.sha1, chunk.encoding )
                chunk.url = self.pkgStore.url( cpath, expiresInSecs )

    def prime( self, localPath, progressCB ):
        """Walk a local directory tree and ensure that all chunks of all files are present in the local cache"""
        self.__storeFiles( self.localCache, localPath, progressCB )

    def validateLocalCache(self):
        return self.__validateStore( self.localCache )

    def validateStore(self):
        return self.__validateStore( self.store )

    def flushLocalCache(self, packageNames ):
        """
        Remove all chunks from the local cache that are not
        referenced by the named packages. Returns the chunks removed.
        """
        packages = [ self.findPackage(pname) for pname in packageNames]
        if len(packages) == 0:
            raise RuntimeError, "flushLocalCache refuses to remove all cached chunks"
        return self.__flushStore( self.localCache, packages )

    def flushStore(self):
        """
        Remote any "dangling" chunks from the store that are not referenced by packages.
        This will happen when packages are deleted Returns the chunks removed.
        """
        packageNames = self.pkgStore.list(TREES_PATH)
        packages = [ self.findPackage(packageName) for packageName in packageNames ]
        return self.__flushStore( self.pkgStore, packages )

    def __validateStore( self, fileStore ):
        """Walk a fileStore and ensure that all chunks are valid sha1 """
        fileList = fileStore.list("")
        corruptedFiles = []
        for fileName in fileList:
            token = fileName.rsplit('/', 2)
            sha1 = token[1] + token[2]
            encoding = package.ENCODING_RAW
            if "zlib" in fileName:
                encoding = package.ENCODING_ZLIB

            buf = fileStore.get(fileName)
            try:
                self.__checkSha1(self.__decompress(buf, encoding), sha1, fileName)
            except:
                corruptedFiles.append({fileName, fileStore.getMetadata(fileName)})
        return corruptedFiles

    def __verifyStore( self, fileStore, pkg ):
        """Walk a fileStore and ensure that all chunks for the given package are present"""
        for pf in pkg.files:
            for chunk in pf.chunks:
                cpath = self.__chunkPath( fileStore, chunk.sha1, chunk.encoding )
                if not fileStore.exists( cpath ):
                    raise RuntimeError, "{0} not found".format(cpath)

    def __flushStore( self, fileStore, packages ):
        """Remove all keys from the store except those referenced by the given packages"""

        # Generate the keys for all chunks in those packages
        keysToKeep = set()
        for pkg in packages:
            for pf in pkg.files:
                for chunk in pf.chunks:
                    keysToKeep.add( (chunk.encoding,chunk.sha1) )

        # Generate the set of all keys currently in the store
        allKeys = set()
        for path in fileStore.list( CHUNKS_PATH ):
            encoding,s1,s2 = fileStore.splitPath(path)
            allKeys.add( (encoding,s1+s2))

        # Work out the keys to remove
        keysToRemove = allKeys.difference( keysToKeep )

        self.outVerbose( "{} packages reference {} chunks...".format(len(packages),len(keysToKeep)))
        self.outVerbose( "The store contains {} chunks, removing {}".format(len(allKeys),len(keysToRemove)))

        # remove them
        if not self.dryRun:
            for encoding,sha1 in keysToRemove:
                fileStore.remove( self.__chunkPath(fileStore, sha1, encoding) )
        return keysToRemove
            
    def __storeFiles( self, store, localPath, progressCB ):
        if not os.path.isdir( localPath ):
            raise IOError( "directory {0} doesn't exist".format( localPath ) )
        packageFiles = []
        for root, dirs, files in os.walk(localPath):
            for file in files:
                rpath = os.path.relpath( os.path.join(root, file), localPath )
                packageFiles.append( self.__storeFile( store, localPath, rpath, progressCB ) )
        return packageFiles

    def __storeFile( self, store, root, rpath, progressCB ):
        filesha1 = hashlib.sha1()
        chunks = []
        self.outVerbose( "Processing file {}", rpath )
        with open( os.path.join( root, rpath ), 'rb' ) as f:
            while True:
                buf = f.read( self.config.chunkSize )
                if not buf:
                    break
                chunksha1 = hashlib.sha1()
                chunksha1.update( buf )
                filesha1.update( buf )
                chunks.append( self.__storeChunk( store, chunksha1.hexdigest(), buf, progressCB ) )
        self.outVerbose( "file {} has hash {}", rpath, filesha1.hexdigest() )
        rpath = package.pathFromFileSystem( rpath )
        return package.PackageFile( filesha1.hexdigest(), rpath, chunks )

    def __storeChunk( self, store, sha1, buf, progressCB ):
        size = len(buf)
        if store.exists( self.__chunkPath( store, sha1, package.ENCODING_RAW ) ):
            progressCB( 0, size )
            return package.FileChunk( sha1, size, package.ENCODING_RAW, None )
        elif store.exists( self.__chunkPath( store, sha1, package.ENCODING_ZLIB ) ):
            progressCB( 0, size )
            return package.FileChunk( sha1, size, package.ENCODING_ZLIB, None )
        else:
            encoding = package.ENCODING_RAW
            if self.config.useCompression:
                buf,encoding = self.__compress( buf )
            if not self.dryRun:
                self.outVerbose( "Uploading {} chunk with hash {}", encoding, sha1  )
                store.put( self.__chunkPath( store, sha1, encoding ), buf )
            progressCB( size, 0 )
            return package.FileChunk( sha1, size, encoding, None )

    def __treeNamePath( self, store, treeName ):
        return store.joinPath( TREES_PATH, treeName )

    def __metaTreeNamePath( self, store, metaTreeName ):
        return store.joinPath( META_TREES_PATH, metaTreeName )

    def __chunkPath( self, store, sha1, encoding ):
        enc = {
            package.ENCODING_RAW : RAW_PATH,
            package.ENCODING_ZLIB : ZLIB_PATH,
        }[encoding]
        return store.joinPath( CHUNKS_PATH, enc, sha1[:2], sha1[2:] )

    def __compress( self, buf ):
        bufz = zlib.compress( buf )
        if len( bufz ) < len( buf ):
            return bufz,package.ENCODING_ZLIB
        else:
            return buf,package.ENCODING_RAW

    def __decompress( self, buf, encoding ):
        if encoding == package.ENCODING_ZLIB:
            return zlib.decompress( buf )
        else:
            return buf

    def __checkSha1( self, buf, sha1, cpath ):
        csha1 = hashlib.sha1()
        csha1.update( buf )
        if csha1.hexdigest() != sha1:
            raise RuntimeError, "sha1 for {0} doesn't match".format( cpath )



import datetime, os

def datetimeFromIso( s ):
    """parse (a subset of ) valid ISO 8601 dates"""
    
    s = s.strip()
    try:
        return datetime.datetime.strptime( s, '%Y-%m-%dT%H:%M:%S.%f' )
    except ValueError:
        return datetime.datetime.strptime( s, '%Y-%m-%dT%H:%M:%S' )
        

def allFilePaths(path):
    """Find all of the files below path"""
    result = []
    for root, dirs, files in os.walk(path):
        for file in files:
            result.append(os.path.normpath(os.path.join(os.path.relpath(root,path), file)))
    return result

def removeEmptyDirectories(path, removeRoot):
  """Remove empty directories recursively"""

  if not os.path.isdir(path):
    return

  files = os.listdir(path)
  if len(files):
    for f in files:
      fullpath = os.path.join(path, f)
      if os.path.isdir(fullpath):
        removeEmptyDirectories(fullpath, True)

  files = os.listdir(path)
  if len(files) == 0 and removeRoot:
    os.rmdir(path)        
    



import os, sys, tempfile, shutil

# You wouldn't think that writing a file would be so hard!

class InPlaceFileWriter(object):
    """
    Just write the file to where it needs to go.
    """
    def __init__(self, filename):
        self.filename = filename

    def __enter__(self):
        self.file = open(self.filename,'wb')
        return self.file        

    def __exit__(self, *args):
        self.file.close()

class PosixAtomicFileWriter(object):
    """
    Rely on posix semantics to update a file atomically,
    by writing to a temp file in the same directory
    and moving it into place
    """
    def __init__(self, filename):
        self.filename = filename

    def __enter__(self):
        self.tempfile = tempfile.NamedTemporaryFile(delete=False,dir=os.path.dirname(self.filename))
        return self.tempfile

    def __exit__(self, *args):
        self.tempfile.close()
        os.rename(self.tempfile.name, self.filename)

class ClunkySemiAtomicWindowsFileWriter(object):
    """
    Windows doesn't have atomic file updates, so
    we try and get close here
    """

    # TODO(timd): consider cutting over to using win32api.ReplaceFile
    # as described here:
    #    http://stupidpythonideas.blogspot.com.au/2014/07/getting-atomic-writes-right.html
    
    def __init__(self, filename):
        self.filename = filename

    def __enter__(self):
        self.tempfile = tempfile.NamedTemporaryFile(delete=False,dir=os.path.dirname(self.filename))
        return self.tempfile

    def __exit__(self, *args):
        # Need to flush both at libc and kernel layers here to
        # ensure that the os.rename() below works correctly under
        # windows
        self.tempfile.flush()
        os.fsync(self.tempfile.fileno())
    
        self.tempfile.close()
        if os.path.exists(self.filename):
           os.unlink(self.filename)
        os.rename(self.tempfile.name, self.filename)

def atomicFileWriter(filename):
    if sys.platform == 'win32':
        return ClunkySemiAtomicWindowsFileWriter(filename)
    else:
        return PosixAtomicFileWriter(filename)



import os

from s3ts.filestore import FileStore

from boto.s3.key import Key
from boto.exception import S3ResponseError

class S3FileStore(FileStore):
    def __init__( self, bucket, pathPrefix=None ):
        self.bucket = bucket
        self.pathPrefix = pathPrefix

    def exists( self, path ):
        k = self._key(path)
        return k.exists()

    def get( self, path ):
        k = self._key(path)
        try:
            return k.get_contents_as_string()
        except S3ResponseError, e:
            if e.status == 404:
                raise KeyError(e)
            raise

    def put( self, path, body ):
        k = self._key(path)
        k.set_contents_from_string( body )

    def list( self, pathPrefix ):
        return [os.path.relpath(key.name,pathPrefix) for key in self.bucket.list(prefix=pathPrefix)]

    def remove( self, path ):
        k = self._key(path)
        k.delete()

    def url( self, path, expiresInSecs ):
        k = self._key(path)
        return k.generate_url(expiresInSecs)

    def joinPath( self, *elements):
        return '/'.join(elements)

    def splitPath(self, path):
        return path.split('/')

    def _key(self,path):
        if self.pathPrefix:
            path = self.joinPath( self.pathPrefix, path )
        return Key(self.bucket,path)



import json, os

from s3ts import filewriter

class FileStore(object):

    def exists( self, path ):
        """Returns true if a file with the given file exists"""
        raise RuntimeError, "Not implemented"

    def get( self, path ):
        """Get the value associated with path

        Raises a KeyError if the path doesn't exist

        """
        raise RuntimeError, "Not implemented"

    def put( self, path, body ):
        """Store a value associated with path

        Overwrites any existing value
        """
        raise RuntimeError, "Not implemented"

    def remove( self, path ):
        """Remove a value with the path, if it exists
        """
        raise RuntimeError, "Not implemented"

    def list( self, pathPrefix ):
        """Return all paths having the specified path prefix"""
        raise RuntimeError, "Not implemented"

    def getFromJson( self, path, jscoder ):
        return jscoder.fromJson( json.loads( self.get(path) ) )

    def putToJson( self, path, v, jscoder ):
        return self.put( path, json.dumps( jscoder.toJson(v) ) )

    def url( self, path, expiresInSecs ):
        raise RuntimeError, "Not implemented"

    def joinPath( self, *elements):
        raise RuntimeError, "Not implemented"

    def splitPath( self, path):
        raise RuntimeError, "Not implemented"

    def getMetadata( self, path):
        raise RuntimeError, "Not implemented"

class FileMetaData:
    def __init__(self,size,lastModified):
        self.size = size
        self.lastModified = lastModified
    def __repr__(self):
        return self.__str__()
    def __str__(self):
        return "size:" + str(self.size) + " mtime:" + str(self.lastModified) 

class LocalFileStore(FileStore):
    """implements the FileStore interface using the local file system"""

    def __init__( self, root ):
        self.root = root

    def __path( self, path ):
        return os.path.join( self.root, path )

    def exists( self, path ):
        return os.path.exists( self.__path(path)  )

    def get( self, path ):
        try:
            with open( self.__path(path), 'rb' ) as f:
                return f.read()
        except IOError, e:
            raise KeyError(e)

    def put( self, path, body ):
        path = self.__path(path)
        dir = os.path.dirname( path )
        if not os.path.isdir( dir ):
            os.makedirs( dir )

        # Do our best to be atomic in our updates here, in case
        # another process is simultaneously updating the file
        with filewriter.atomicFileWriter(path) as f:
            f.write(body)

    def remove( self, path ):
        path = self.__path(path)
        if os.path.exists( path ):
            os.unlink(path)

    def list( self, pathPrefix ):
        results = []
        for dir0, dirs, files in os.walk(self.__path(pathPrefix)):
            for file in files:
                path = os.path.join( dir0, file)
                rpath = os.path.relpath( path, os.path.join( self.root, pathPrefix ) )
                results.append( rpath )
        return results
    
    def joinPath( self, *elements):
        return os.path.join(*elements)

    def splitPath( self, path):
        return path.split(os.sep)

    def getMetadata(self, path):
        """Returns the size and update time for the given path
           Raises a KeyError if the path doesn't exist
        """
        statinfo = os.stat(self.__path(path))
        return FileMetaData(statinfo.st_size, statinfo.st_mtime)






"""
contains classes for configuration and settings
"""

import os, json, datetime

from s3ts.utils import datetimeFromIso

class S3Access(object):
    """Provide connection parameters for an s3m store"""

    def __init__( self, awsKey, awsSecret, s3Bucket ):
        self.awsKey = awsKey
        self.awsSecret = awsSecret
        self.s3Bucket = s3Bucket

        
class S3AccessJS(object):
    """De/Serialise S3Access objects"""
    
    def fromJson( self, jv ):
        return S3Access(
            jv['awsKey'],
            jv['awsSecret'],
            jv['s3Bucket']
        )
    
class TreeStoreConfig(object):
    """Configuration data for an s3m store"""

    def __init__( self, chunkSize, useCompression ):
        self.chunkSize = chunkSize
        self.useCompression = useCompression

        
class TreeStoreConfigJS(object):
    """De/serialise Config objects"""

    def fromJson( self, jv ):
        return TreeStoreConfig(
            jv['chunkSize'],
            jv['useCompression']
            )

    def toJson( self, v ):
        return {
            'chunkSize' : v.chunkSize,
            'useCompression' : v.useCompression,
            }

class InstallProperties(object):
    """records the details of an installation"""

    def __init__( self, treeName, installTime ):
        self.treeName = treeName
        self.installTime = installTime

class InstallPropertiesJS(object):
    """De/serialise InstallProperties objects"""

    def fromJson( self, jv ):
        return InstallProperties(
            jv['treeName'],
            datetimeFromIso( jv['installTime'] )
            )

    def toJson( self, v ):
        return {
            'treeName' : v.treeName,
            'installTime' : v.installTime.isoformat(),
            }

S3TS_PROPERTIES = '.s3ts.properties'    


def writeInstallProperties( installDir, props ):
    with open( os.path.join( installDir, S3TS_PROPERTIES ), 'w' ) as f:
        f.write( json.dumps( InstallPropertiesJS().toJson( props ) ) )

def readInstallProperties( installDir ):
    with open( os.path.join( installDir, S3TS_PROPERTIES ), 'r' ) as f:
        return InstallPropertiesJS().fromJson( json.loads( f.read() ) )
    

        



from distutils.core import setup

setup(
    name="s3ts",

    version="0.1.0",

    author="Tim Docker",
    author_email="timd@helixta.com.au",

    # Contents
    packages=["s3ts"],
    package_dir={'' : 'src'},

    include_package_data=False,

    # Details
    url="https://bitbucket.org/helix-collective/s3ts/wiki/Home",

    description="A library to manage versioned tree based data on S3",

    install_requires=[
        "boto",
        "requests"
    ],
)



import os, tempfile, unittest, shutil, subprocess, datetime, time

from s3ts.filestore import LocalFileStore
from s3ts.s3filestore import S3FileStore
from s3ts.config import TreeStoreConfig, readInstallProperties, S3TS_PROPERTIES
from s3ts.treestore import TreeStore
from s3ts.utils import datetimeFromIso
from s3ts.package import PackageJS, S3TS_PACKAGEFILE
from s3ts.metapackage import MetaPackage, SubPackage

import boto
import logging

# boto.set_stream_logger('boto')

class CaptureDownloadProgress:
    def __init__( self ):
        self.recorded = []

    def __call__( self, bytesDownloaded, bytesFromCache ):
        self.recorded.append( bytesDownloaded + bytesFromCache )

CaptureUploadProgress = CaptureDownloadProgress

class CaptureInstallProgress:
    def __init__( self ):
        self.recorded = []

    def __call__( self, nBytes ):
        self.recorded.append( nBytes )


class EmptyS3Bucket:
    def __init__( self, bucket ):
        self.bucket = bucket

    def __enter__(self):
        # Ensure the bucket starts empty
        assert len(list(self.bucket.list()))==0, "S3 bucket is not empty"

    def __exit__(self, type, value, traceback):
        # Clean the bucket (ok, as we know it started empty)
        self.bucket.delete_keys( self.bucket.list() )


class TestTreeStore(unittest.TestCase):

    def setUp(self):
        self.workdir = tempfile.mkdtemp()
        if os.path.exists( self.workdir ):
            shutil.rmtree( self.workdir )
        os.makedirs( self.workdir )

        self.FILE1 = '#!/bin/env python\n def main(): print "hello"\n'
        self.FILE2 = '#!/bin/env python\n def main(): print "goodbye"\n'
        self.FILE2_A = '#!/bin/env python\n def main(): print "goodbye foreever"\n'
        self.FILE3 = '#!/bin/env python\n def main(): print "goodbye foreever"\n'
        self.FILE4 = '#!/bin/env python\n def main(): print "what now"\n' 
        self.FILE5 = 'Just text' 
        self.CAR01 = (
            'Some big and complicated data structure goes here, hopefully big enough that it requires chunking and compression.\n'
            'sydney london paris port moresby okinawa st petersburg salt lake city  new york whitehorse mawson woy woy st louis\n'
            )
        

        # Create some test input data
        self.srcTree = makeEmptyDir( os.path.join( self.workdir, 'src-1' ) )
        fs = LocalFileStore( self.srcTree )
        fs.put( 'code/file1.py', self.FILE1)
        fs.put( 'code/file2.py', self.FILE2)
        fs.put( 'assets/car-01.db', self.CAR01)

        self.srcTree2 = makeEmptyDir( os.path.join( self.workdir, 'src-2' ) )
        fs = LocalFileStore( self.srcTree2 )
        fs.put( 'code/file1.py', self.FILE1 )
        fs.put( 'code/file3.py', self.FILE3 )
        fs.put( 'code/file4.py', self.FILE4)
        fs.put( 'assets/car-01.db', self.CAR01 )

        self.srcTree3 = makeEmptyDir( os.path.join( self.workdir, 'src-3' ) )
        fs = LocalFileStore( self.srcTree3 )
        fs.put( 'code/file1.py', self.FILE1 )
        fs.put( 'code/file2.py', self.FILE2_A )
        fs.put( 'code/file4.py', self.FILE4 )
        fs.put( 'text/text', self.FILE5 )

        self.srcTree4 = makeEmptyDir( os.path.join( self.workdir, 'src-4' ) )
        fs = LocalFileStore( self.srcTree4 )
        fs.put( 'file1.py', self.FILE1 )
        fs.put( 'code/file2.py', self.FILE2_A )
        fs.put( 'code/file4.py', self.FILE4 )
        fs.put( 'text', self.FILE5 )

        self.srcVariant = makeEmptyDir( os.path.join( self.workdir, 'src1-kiosk' ) )
        fs = LocalFileStore( self.srcVariant )
        fs.put( 'kiosk-01/key', 'this is the key src1:kiosk-01' )
        fs.put( 'kiosk-02/key', 'this is the key src1:kiosk-02' )

    def tearDown(self):
        shutil.rmtree( self.workdir )

    def test_fs_treestore(self):
        # Create a file system backed treestore
        fileStore = LocalFileStore( makeEmptyDir( os.path.join( self.workdir, 'fs' ) ) )
        localCache = LocalFileStore( makeEmptyDir( os.path.join( self.workdir, 'cache' ) ) )
        treestore = TreeStore.create( fileStore, localCache, TreeStoreConfig( 100, True ) )

        # Upload 2 trees
        creationTime = datetimeFromIso( '2015-01-01T00:00:00.0' )
        treestore.upload( 'v1.0', '', creationTime, self.srcTree, CaptureUploadProgress() )
        pkg = treestore.findPackage( 'v1.0' )

        # Confirm it's in the index
        self.assertEquals( treestore.listPackages(), ['v1.0'] )

        # Verify it
        treestore.verify( pkg )

        # Test the cache priming function
        treestore.prime( self.srcTree2, CaptureUploadProgress() )

        # Test whether the verifyCache works
        corruptedFiles = treestore.validateLocalCache()
        self.assertEquals( len(corruptedFiles), 0) 
        
        # Download it, checking we get expected progress callbacks
        # The order of the callbacks will depend on the order of the
        # chunks in the package definition, which will depend on the
        # iteration order of the file system when the package was created.
        # So check independently of ordering.
        cb = CaptureDownloadProgress()
        treestore.download( pkg, cb )
        self.assertEquals( sorted(cb.recorded), [30, 45, 47, 100, 100] )

        # Verify it locally
        treestore.verifyLocal( pkg )

        # Install it
        destTree = os.path.join( self.workdir, 'dest-1' )
        treestore.install( pkg, destTree, CaptureInstallProgress() )

        # Check that the installed tree is the same as the source tree
        self.assertEquals( subprocess.call( 'diff -r -x {0} {1} {2}'.format(S3TS_PROPERTIES,self.srcTree,destTree), shell=True ), 0 )

        # Rename the tree, and check that installing that is the same
        treestore.rename( 'v1.0', 'v1.0x' )
        pkg = treestore.findPackage( 'v1.0x' )
        treestore.download( pkg, CaptureDownloadProgress() )
        destTree = os.path.join( self.workdir, 'dest-2' )
        treestore.install( pkg, destTree, CaptureInstallProgress() )
        self.assertEquals( subprocess.call( 'diff -r -x {0} {1} {2}'.format(S3TS_PROPERTIES,self.srcTree,destTree), shell=True ), 0 )

        # Test the flushStore function has nothing to remove)
        treestore.upload( 'extra', '', creationTime, self.srcTree2, CaptureUploadProgress() )
        removed = treestore.flushStore()
        self.assertEquals(len(removed), 0)

        # Remove a tree
        treestore.remove( 'v1.0x' )

        # Test the store now has dangling chunks when can be removed
        removed = treestore.flushStore()
        self.assertTrue(len(removed) > 0)

        treestore.upload( 'v1.0', '', creationTime, self.srcTree, CaptureUploadProgress() )

        # Initially the local cache should contain chunks for v1.0 and extra. Empty
        # the local cache by successive flush operations
        removed = treestore.flushLocalCache(['extra'])
        self.assertTrue(len(removed) > 0)
        removed = treestore.flushLocalCache(['v1.0'])
        self.assertTrue(len(removed) > 0)

        # Confirm that removing everything from the local cache is refused
        with self.assertRaises(RuntimeError):
            treestore.flushLocalCache([])

    def test_sync(self):            
        # Create a file system backed treestore
        fileStore = LocalFileStore( makeEmptyDir( os.path.join( self.workdir, 'fs' ) ) )
        localCache = LocalFileStore( makeEmptyDir( os.path.join( self.workdir, 'cache' ) ) )
        treestore = TreeStore.create( fileStore, localCache, TreeStoreConfig( 10, True ) )

        creationTime = datetimeFromIso( '2015-01-01T00:00:00.0' )
        treestore.upload( 'v1.0', '', creationTime, self.srcTree, CaptureUploadProgress() )
        treestore.upload( 'v1.3', '', creationTime, self.srcTree3, CaptureUploadProgress() )
        treestore.upload( 'v1.4', '', creationTime, self.srcTree4, CaptureUploadProgress() )

        testdir = makeEmptyDir( os.path.join( self.workdir, 'test' ) )

        def assertExists( path ):
            self.assertTrue( os.path.exists( os.path.join(testdir, path) ) )

        def assertContains( path, text ):
            self.assertEquals( open( os.path.join(testdir, path) ).read(), text )

        def assertDoesntExist( path ):
            self.assertFalse( os.path.exists( os.path.join(testdir, path) ) )

        def assertInstalled(pkg, testdir):
            result = treestore.compareInstall(pkg, testdir)
            self.assertEquals( result.missing, set() )
            self.assertEquals( result.extra, set() )
            self.assertEquals( result.diffs, set() )
        
        # sync a package to an empty directory
        pkg = treestore.findPackage('v1.0')
        treestore.download( pkg, CaptureDownloadProgress() ) 
        treestore.sync( pkg, testdir, CaptureInstallProgress() )
        assertContains( "code/file1.py", self.FILE1 )
        assertContains( "code/file2.py", self.FILE2 )
        assertContains( "assets/car-01.db", self.CAR01 )
        assertExists( S3TS_PACKAGEFILE )
        assertInstalled( pkg, testdir )

        # Re-sync the same package
        pkg = treestore.findPackage('v1.0')
        treestore.download( pkg, CaptureDownloadProgress() ) 
        treestore.sync( pkg, testdir, CaptureInstallProgress() )
        assertContains( "code/file1.py", self.FILE1 )
        assertContains( "code/file2.py", self.FILE2 )
        assertContains( "assets/car-01.db", self.CAR01 )
        assertExists( S3TS_PACKAGEFILE )
        assertInstalled( pkg, testdir )

        # Sync to a different package
        pkg = treestore.findPackage('v1.3')
        treestore.download( pkg, CaptureDownloadProgress() ) 
        treestore.sync( pkg, testdir, CaptureInstallProgress() )
        assertContains( "code/file1.py", self.FILE1 )
        assertContains( "code/file2.py", self.FILE2_A )
        assertDoesntExist( "assets/car-01.db" )
        assertContains( "code/file4.py", self.FILE4 )
        assertContains( "text/text", self.FILE5 )
        assertExists( S3TS_PACKAGEFILE )
        assertInstalled( pkg, testdir )

        # Sync back to the first package
        pkg = treestore.findPackage('v1.0')
        treestore.download( pkg, CaptureDownloadProgress() ) 
        treestore.sync( pkg, testdir, CaptureInstallProgress() )
        assertContains( "code/file1.py", self.FILE1 )
        assertContains( "code/file2.py", self.FILE2 )
        assertContains( "assets/car-01.db", self.CAR01 )
        assertDoesntExist( "code/file4.py" )
        assertExists( S3TS_PACKAGEFILE )
        assertInstalled( pkg, testdir )

        # Remove the package file, and sync the second package again
        os.unlink( os.path.join( testdir, S3TS_PACKAGEFILE ) )
        pkg = treestore.findPackage('v1.3')
        treestore.download( pkg, CaptureDownloadProgress() ) 
        treestore.sync( pkg, testdir, CaptureInstallProgress() )
        assertContains( "code/file1.py", self.FILE1 )
        assertContains( "code/file2.py", self.FILE2_A )
        assertDoesntExist( "assets/car-01.db" )
        assertContains( "code/file4.py", self.FILE4 )
        assertExists( S3TS_PACKAGEFILE )
        assertInstalled( pkg, testdir )

        # Add an extra file not in the package, and ensure
        # that syncing deletes it
        with open( os.path.join(testdir, "debug.log"), 'w') as f:
            f.write( "something" )
        pkg = treestore.findPackage('v1.3')
        treestore.sync( pkg, testdir, CaptureInstallProgress() )
        assertInstalled( pkg, testdir )

        # Sync to test replacing a directory with a file
        pkg = treestore.findPackage('v1.4')
        treestore.download( pkg, CaptureDownloadProgress() ) 
        treestore.sync( pkg, testdir, CaptureInstallProgress() )
        assertContains( "text", self.FILE5 )
        assertInstalled( pkg, testdir )

    def test_metapackages(self):
        # Create a file system backed treestore
        fileStore = LocalFileStore( makeEmptyDir( os.path.join( self.workdir, 'fs' ) ) )
        localCache = LocalFileStore( makeEmptyDir( os.path.join( self.workdir, 'cache' ) ) )
        treestore = TreeStore.create( fileStore, localCache, TreeStoreConfig( 10, True ) )
        
        creationTime = datetimeFromIso( '2015-01-01T00:00:00.0' )
        treestore.upload( 'v1.0', '', creationTime, self.srcTree, CaptureUploadProgress() )
        treestore.upload( 'v1.3', '', creationTime, self.srcTree3, CaptureUploadProgress() )
        treestore.upload( 'v1.4', '', creationTime, self.srcTree4, CaptureUploadProgress() )

        meta1 = MetaPackage(
            name = 'meta1',
            description = '',
            creationTime = creationTime,
            components = [
                SubPackage( 'dir-1', 'v1.0' ),
                SubPackage( 'dir-2', 'v1.3' ),
            ]
        )

        meta1.verify(treestore,{})
        treestore.uploadMetaPackage(meta1)
        meta1p = treestore.find( 'meta1', {})
        treestore.download(meta1p, CaptureDownloadProgress() )

        # Install it
        destTree = os.path.join( self.workdir, 'dest-1' )
        treestore.install(meta1p, destTree, CaptureInstallProgress() )
        def assertContains( path, text ):
            self.assertEquals( open( os.path.join(destTree, path) ).read(), text )
        assertContains("dir-1/code/file1.py", self.FILE1)
        assertContains("dir-2/text/text", self.FILE5)
                
    def test_s3_treestore(self):
        # Create an s3 backed treestore
        # Requires these environment variables set
        #
        #   AWS_ACCESS_KEY_ID
        #   AWS_SECRET_ACCESS_KEY
        #   S3TS_BUCKET
        #
        # NB: **this will only work if the bucket is empty

        s3c = boto.connect_s3()
        bucket = s3c.get_bucket( os.environ['S3TS_BUCKET'] )

        with EmptyS3Bucket(bucket):
            fileStore = S3FileStore( bucket )
            localCache = LocalFileStore( makeEmptyDir( os.path.join( self.workdir, 'cache' ) ) )
            treestore = TreeStore.create( fileStore, localCache, TreeStoreConfig( 100, True ) )

            # Upload it as a tree
            creationTime = datetimeFromIso( '2015-01-01T00:00:00.0' )
            treestore.upload( 'v1.0', '', creationTime, self.srcTree, CaptureUploadProgress() )
            pkg = treestore.findPackage( 'v1.0' )

            # Confirm it's in the index
            self.assertEquals( treestore.listPackages(), ['v1.0'] )

            # Verify it
            treestore.verify( pkg )

            # Download it, checking we get expected progress callbacks
            cb = CaptureDownloadProgress()
            treestore.download( pkg, cb )
            self.assertEquals( sorted(cb.recorded), [30, 45, 47, 100, 100] )

            # Verify it locally
            treestore.verifyLocal( pkg )

            # Install it
            destTree = os.path.join( self.workdir, 'dest-1' )
            treestore.install( pkg, destTree, CaptureInstallProgress() )

            # Check that the installed tree is the same as the source tree
            self.assertEquals( subprocess.call( 'diff -r -x {0} {1} {2}'.format(S3TS_PROPERTIES,self.srcTree,destTree), shell=True ), 0 )
            self.assertEquals( readInstallProperties(destTree).treeName, 'v1.0' )

            # Use the compareInstall function to confirm the installed package is ok, and
            # then check that modifying the files show up in the comparison
            result = treestore.compareInstall( pkg, destTree )
            self.assertEquals( len(result.missing), 0 )
            self.assertEquals( len(result.extra), 0 )
            self.assertEquals( len(result.diffs), 0 )

            with open( os.path.join(destTree,"code/file1.py"), "w" ) as f:
                f.write("x")
            with open( os.path.join(destTree,"code/file3.py"), "w" ) as f:
                f.write("y")
            os.unlink(os.path.join(destTree,'assets/car-01.db'))
            
            result = treestore.compareInstall( pkg, destTree )
            self.assertEquals( result.missing, set(['assets/car-01.db']) )
            self.assertEquals( result.extra, set(['code/file3.py']) )
            self.assertEquals( result.diffs, set(['code/file1.py']) )

            # Reinstall to fix directory content
            shutil.rmtree( destTree )
            treestore.install( pkg, destTree, CaptureInstallProgress() )
            result = treestore.compareInstall( pkg, destTree )
            self.assertEquals( len(result.missing), 0 )
            self.assertEquals( len(result.extra), 0 )
            self.assertEquals( len(result.diffs), 0 )

            # Now create a pre-signed version of the package
            pkg = treestore.findPackage( 'v1.0' )
            treestore.addUrls( pkg, 3600 )
            self.assertEquals( len(result.missing), 0 )
            self.assertEquals( len(result.extra), 0 )
            self.assertEquals( len(result.diffs), 0 )

            # And download it directly via http. Create a new local cache
            # to ensure that we actually redownload each chunk
            localCache = LocalFileStore( makeEmptyDir( os.path.join( self.workdir, 'cache' ) ) )
            treestore2 = TreeStore.forHttpOnly( localCache )
            cb = CaptureDownloadProgress()
            treestore2.downloadHttp( pkg, cb )
            self.assertEquals( sorted(cb.recorded), [30, 45, 47, 100, 100] )

            # Install it
            destTree2 = os.path.join( self.workdir, 'dest-2' )
            treestore2.install( pkg, destTree2, CaptureInstallProgress() )

            # Check that the new installed tree is the same as the source tree
            self.assertEquals( subprocess.call( 'diff -r -x {0} {1} {2}'.format(S3TS_PROPERTIES,self.srcTree,destTree2), shell=True ), 0 )

            # Rename the tree, and check that installing that is the same
            treestore.rename( 'v1.0', 'v1.0x' )
            pkg = treestore.findPackage( 'v1.0x' )
            treestore.download( pkg, CaptureDownloadProgress() )
            destTree = os.path.join( self.workdir, 'dest-3' )
            treestore.install( pkg, destTree, CaptureInstallProgress() )
            self.assertEquals( subprocess.call( 'diff -r -x {0} {1} {2}'.format(S3TS_PROPERTIES,self.srcTree,destTree), shell=True ), 0 )

            # Remove the tree
            treestore.remove( 'v1.0x' )
            
    def test_s3_prefixes(self):
        # Requires these environment variables set
        #
        #   AWS_ACCESS_KEY_ID
        #   AWS_SECRET_ACCESS_KEY
        #   S3TS_BUCKET
        #
        # NB: **this will only work if the bucket is empty

        s3c = boto.connect_s3()
        bucket = s3c.get_bucket( os.environ['S3TS_BUCKET'] )

        with EmptyS3Bucket(bucket):
            localCache = LocalFileStore( makeEmptyDir( os.path.join( self.workdir, 'cache' ) ) )
            treestore1 = TreeStore.create( S3FileStore( bucket, "prefix1" ), localCache, TreeStoreConfig( 100, True ) )
            treestore2 = TreeStore.create( S3FileStore( bucket, "prefix2" ), localCache, TreeStoreConfig( 100, True ) )

            # Confirm we can write the different values to the same path in both treestores,
            # and the different prefix keeps them separate independent
            creationTime = datetimeFromIso( '2015-01-01T00:00:00.0' )
            treestore1.upload( 'release', '', creationTime, self.srcTree, CaptureUploadProgress() )
            treestore2.upload( 'release', '', creationTime, self.srcTree2, CaptureUploadProgress() )
            pkg1 = treestore1.findPackage( 'release' )
            pkg2 = treestore2.findPackage( 'release' )
            self.assertEquals(len(pkg1.files),3)
            self.assertEquals(len(pkg2.files),4)

    def test_s3_merged_package(self):
        # Test the creation and subsequent installation of merged packages
        # Requires these environment variables set
        #
        #   AWS_ACCESS_KEY_ID
        #   AWS_SECRET_ACCESS_KEY
        #   S3TS_BUCKET
        #
        # NB: **this will only work if the bucket is empty

        s3c = boto.connect_s3()
        bucket = s3c.get_bucket( os.environ['S3TS_BUCKET'] )

        with EmptyS3Bucket(bucket):
            localCache = LocalFileStore( makeEmptyDir( os.path.join( self.workdir, 'cache' ) ) )
            treestore = TreeStore.create( S3FileStore( bucket), localCache, TreeStoreConfig( 100, True ) )
            creationTime = datetimeFromIso( '2015-01-01T00:00:00.0' )
            treestore.upload( 'src1', '', creationTime, self.srcTree, CaptureUploadProgress() )
            treestore.upload( 'src2', '', creationTime, self.srcTree2, CaptureUploadProgress() )
            treestore.upload( 'src3', '', creationTime, self.srcTree3, CaptureUploadProgress() )
            treestore.createMerged( 'merged', creationTime, { '.' : 'src1', 'subdir-a' : 'src2', 'subdir-b' : 'src3'})
            pkg = treestore.findPackage( 'merged' )
            treestore.download( pkg, CaptureDownloadProgress() )
            destTree = os.path.join( self.workdir, 'merged' )
            treestore.install( pkg, destTree, CaptureInstallProgress() )

            def assertSameContent( path1, path2 ):
                with open(path1) as f1:
                    with open(path2) as f2:
                        self.assertEquals( f1.read(), f2.read() )

            assertSameContent(os.path.join(destTree, "code/file1.py"), os.path.join(self.srcTree, "code/file1.py"))
            assertSameContent(os.path.join(destTree, "subdir-a/code/file4.py"), os.path.join(self.srcTree2, "code/file4.py"))
            assertSameContent(os.path.join(destTree, "subdir-b/text/text"), os.path.join(self.srcTree3, "text/text"))

    def test_s3_many_treestore(self):
        # Create an s3 backed treestore
        # Requires these environment variables set
        #
        #   AWS_ACCESS_KEY_ID
        #   AWS_SECRET_ACCESS_KEY
        #   S3TS_BUCKET
        #
        # NB: **this will only work if the bucket is empty

        s3c = boto.connect_s3()
        bucket = s3c.get_bucket( os.environ['S3TS_BUCKET'] )

        with EmptyS3Bucket(bucket):
            fileStore = S3FileStore( bucket )
            localCache = LocalFileStore( makeEmptyDir( os.path.join( self.workdir, 'cache' ) ) )
            treestore = TreeStore.create( fileStore, localCache, TreeStoreConfig( 100, True ) )

            # Upload it as a tree
            creationTime = datetimeFromIso( '2015-01-01T00:00:00.0' )
            treestore.uploadMany( 'v1.0', '', creationTime, self.srcTree, self.srcVariant, CaptureUploadProgress() )
            pkg = treestore.findPackage( 'v1.0:kiosk-01' )

            # Confirm it's in the index
            self.assertEquals( treestore.listPackages(), ['v1.0:kiosk-01', 'v1.0:kiosk-02'] )

            # Verify it
            treestore.verify( pkg )

            # Download it, checking we get expected progress callbacks
            cb = CaptureDownloadProgress()
            treestore.download( pkg, cb )
            self.assertEquals( sorted(cb.recorded), [29, 30, 45, 47, 100, 100] )

            # Verify it locally
            treestore.verifyLocal( pkg )

            # Install it
            destTree = os.path.join( self.workdir, 'dest-1' )
            treestore.install( pkg, destTree, CaptureInstallProgress() )

            # Check that the installed tree is the same as the source tree
            self.assertEquals( subprocess.call( 'diff -r -x {0} {1} {2}'.format(S3TS_PROPERTIES,self.srcTree + '/assets',destTree + '/assets'), shell=True ), 0 )
            self.assertEquals( subprocess.call( 'diff -r -x {0} {1} {2}'.format(S3TS_PROPERTIES,self.srcTree + '/code',destTree + '/code'), shell=True ), 0 )


            self.assertEquals( readInstallProperties(destTree).treeName, 'v1.0:kiosk-01' )



def makeEmptyDir( path ):
    if os.path.exists( path ):
        shutil.rmtree( path )
    os.makedirs( path )
    return path



if __name__ == '__main__':
    unittest.main()



import json, datetime, os
from collections import OrderedDict

from s3ts.utils import datetimeFromIso
from s3ts import package

class MetaPackage(object):
    """
    A metapackage specified how multiple packages can be
    arranged into a single directory tree.
    """

    def __init__( self, name, description, creationTime, components ):
        self.name = name
        self.description = description
        self.creationTime = creationTime
        self.components = components

    def verify(self, treestore, metadata):
        assert self.name
        for component in self.components:
            component.verify(treestore, metadata)

    def package(self, treestore, metadata):
        files = []
        for component in self.components:
            files += component.packageFiles(treestore, metadata)
        return package.Package( self.name, self.description, self.creationTime, files )
        
class SubPackage(object):
    """
    A sub package component of a MetaPackage.
    """
    def __init__( self, installPath, packageName):
        self.installPath = installPath
        self.packageName = packageName

    def verify(self, treestore, metadata):
        assert self.installPath
        assert self.packageName
        try:
            treestore.findPackage(self.packageName)
        except KeyError:
            raise KeyError, "package {} doesn't exist".format(self.packageName)

    def packageFiles(self, treestore, metadata):
        subpackage = treestore.findPackage(self.packageName)
        return packageFilesInSubdir(self.installPath, subpackage)

    def info(self):
        return '{} -> package {}'.format( self.installPath, self.packageName )

class LocalizedPackage(object):        
    """
    A sub package component of a MetaPackage, with localised installation.

    The localizedPackageName may use metadata keys , eg `local-{hostname}`. In
    which case the hostname metadata value will be used to generate the 
    actual package to be installed. If a package with this name is not available,
    then the sepewill be installed.

    """
    def __init__( self, installPath, localizedPackageName, defaultPackageName):
        self.installPath = installPath
        self.localizedPackageName = localizedPackageName
        self.defaultPackageName = defaultPackageName

    def verify(self, treestore, metadata):
        assert self.installPath
        assert self.localizedPackageName
        assert self.defaultPackageName
        metakeys = re.re.findall( '{([^}]*)}', self.localizedPackageName)
        for key in metakeys:
            if not metadata.has_key(key):
                raise RuntimeError, "Package references unknown metdata key '{}'".format(key)
            packageName = self.localizedPackageName.format(**metdata)
            try:
                treestore.findPackage(self.packageName)
            except KeyError:
                treestore.findPackage(defaultPackageName)

    def packageFiles(self, treestore, metadata):
        packageName = self.localizedPackageName.format(**metdata)
        try:
            subpackage = treestore.get(packageName)
        except KeyError:
            package = treestore.get(self.defaultPackageName)
        return packageFilesInSubdir(self.installPath, subpackage)

    def info(self):
        return '{} -> package {}'.format(self.installPath, self.localizedPackageName)

# In the json serialization code below we use OrderedDict
# to lock down the order of the fields in the generated json.
# We do this because we intend the resulting json to
# be human readable.

class MetaPackageJS(object):
    """A json de/serialiser for MetaPackage objects"""
    def __init__( self ):
        self.componentJS = ComponentJS()

    def fromJson( self, jv ):
        return MetaPackage(
            jv['name'],
            jv.get('description', ''),
            datetimeFromIso( jv['creationTime'] ),
            [self.componentJS.fromJson(jv1) for jv1 in jv['components']]
        )
                        
    def toJson( self, v ):
        return OrderedDict([
            ('name', v.name),
            ('description', v.description),
            ('creationTime', v.creationTime.isoformat()),
            ('components', [ self.componentJS.toJson(f) for f in v.components ]),
        ])

class ComponentJS(object):
    """A json de/serialiser for MetaPackage Component objects"""
    def __init__( self ):
        self.subPackageJS = SubPackageJS()
        self.localizedPackageJS = LocalizedPackageJS()

    def fromJson( self, jv ):
        if jv.has_key('subPackage'):
            return self.subPackageJS.fromJson( jv['subPackage'] )
        elif jv.has_key('localizedPackage'):
            return self.localizedPackageJS.fromJson( jv['localizedPackage'] )
        else:
            raise RuntimeError, "invalid metapackage component"
        
    def toJson( self, v ):
        if isinstance( v, SubPackage ):
            return { 'subPackage' : self.subPackageJS.toJson(v) }
        elif isinstance( v, LocalizedPackage ):
            return { 'localizedPackage' : self.localizedPackageJS.toJson(v) }
        else:
            raise RuntimeError, "invalid metapackage component"

class SubPackageJS(object):
    """A json de/serializer for metapackage SubPackage objects"""

    def fromJson( self, jv ):
        return SubPackage( jv['installPath'], jv['packageName'] )

    def toJson( self, v ):
        return OrderedDict([
            ('installPath', v.installPath),
            ('packageName', v.packageName),
        ])

class LocalizedPackageJS(object):
    """A json de/serializer for metapackage LocalizedPackage objects"""

    def fromJson( self, jv ):
        return LocalizedPackage( jv['installPath'], jv['localizedPackageName'], jv['defaultPackageName'] )

    def toJson( self, v ):
        return OrderedDict([
            ('installPath', v.installPath),
            ('localizedPackageName', v.localizedPackageName),
            ('defaultPackageName', v.defaultPackageName),
        ])
    
def packageFilesInSubdir(subdir, subpackage):
    files = []
    for file in subpackage.files:
        path = package.pathFromFileSystem( os.path.normpath( os.path.join(subdir, file.path) ) )
        files.append( package.PackageFile( file.sha1, path, file.chunks ) )
    return files
        
                
